@@ -3,8 +3,6 @@ from clusterval.internal import calculate_internal
 from clusterval.evaluate import Clusterval
 
 
-def __doc__():
-    return "This is help"
 
 
 
